/*
 *	filter.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <sp/sp.h>
#include <sp/base.h>
#include <sp/memory.h>
#include <sp/voperate.h>
#include <sp/fft.h>
#include <sp/kaiser.h>
#include <sp/filter.h>

void dvangle(DVector x)
{
    long k;

    if (x->imag == NULL) {
	dvzeros(x, x->length);
	return;
    }

    for (k = 0; k < x->length; k++) {
	if (x->data[k] == 0.0 && x->imag[k] == 0.0) {
	    x->data[k] = 0.0;
	} else {
	    x->data[k] = atan2(x->imag[k], x->data[k]);
	}
    }
    dvreal(x);

    return;
}

DVector xdvangle(DVector x)
{
    long k;
    DVector phs;

    if (x->imag == NULL) {
	phs = xdvzeros(x->length);
	return phs;
    } else {
	phs = xdvalloc(x->length);
    }

    for (k = 0; k < phs->length; k++) {
	if (x->data[k] == 0.0 && x->imag[k] == 0.0) {
	    phs->data[k] = 0.0;
	} else {
	    phs->data[k] = atan2(x->imag[k], x->data[k]);
	}
    }

    return phs;
}

void dvunwrap(DVector phs, double cutoff)
{
    long k;
    double a;
    DVector f;
    
    if (cutoff <= 0.0) {
	cutoff = PI;
    }

    a = dvmin(phs, NULL);
    for (k = 0; k < phs->length; k++) {
	phs->data[k] = rem(phs->data[k] - a, 2.0 * PI) + a;
    }

    f = xdvzeros(phs->length);
    for (k = 1; k < f->length; k++) {
	a = phs->data[k] - phs->data[k - 1];
	if (a > cutoff) {
	    f->data[k] = -2.0 * PI;
	} else if (a < -cutoff) {
	    f->data[k] = 2.0 * PI;
	}
    }

    dvcumsum(f);
    dvoper(phs, "+", f);

    xdvfree(f);

    return;
}

double sinc(double x)
{
    double a;

    if (x == 0.0) {
	a = 1.0;
    } else {
	a = sin(x) / x;
    }
    return a;
}

double sincc(double x, double c)
{
    double a;

    if (x == 0.0) {
	a = c;
    } else {
	a = (c * sin(x)) / x;
    }
    return a;
}

DVector xdvlowpass(double cutoff, double sidelobe, double trans, double gain)
{
    long k;
    long half, length;
    double beta;
    double value;
    DVector filter;

    /* get kaiser parameter */
    getkaiserparam(sidelobe, trans, &beta, &length);

    /* make sure length is odd */
    length = (long)(length / 2) * 2 + 1;
    half = (length - 1) / 2;

    spDebug(10, "xdvlowpass", "cutoff = %f, gain = %f, beta = %f, length = %ld\n",
	    cutoff, gain, beta, length);

    /* memory allocate */
    filter = xdvalloc(length);

    /* get kaiser window */
    kaiser(filter->data, filter->length, beta);

    /* calculate lowpass filter */
    for (k = 0; k < length; k++) {
	value = sincc(PI * cutoff * (double)(k - half), gain * cutoff);
	filter->data[k] *= value;
    }

    return filter;
}

#define getnumloop(x,y) MAX((int)(((x) + (y) - 1) / (y)), 1)

DVector xdvfftfiltm(DVector b, DVector x, long fftl)
{
    int fftl_p;
    long i, nloop;
    long k;
    long pos = 0;
    long length, block;
    double real, imag;
    double *hr, *hi;
    double *xr, *xi;
    DVector xo;

    /* get fft point */
    if (x->length < b->length) {
	length = b->length + x->length - 1;
    } else {
	length = b->length;
    }
    fftl_p = nextpow2(MAX(fftl, 2 * length));
    fftl = POW2(fftl_p);
	
    /* initialize */
    length = b->length / 2 - 1;
    block = fftl - b->length + 1;
    nloop = getnumloop(x->length, block);

    /* memory allocate */
    hr = xalloc(fftl, double);
    hi = xalloc(fftl, double);
    xr = xalloc(fftl, double);
    xi = xalloc(fftl, double);
    if (x->imag != NULL || b->imag != NULL) {
	xo = xdvrizeros(x->length);
    } else {
	xo = xdvzeros(x->length);
    }

    /* fft of the filter */
    for (k = 0; k < fftl; k++) {
	if (k < b->length) {
	    hr[k] = b->data[k];
	    if (b->imag != NULL) {
		hi[k] = b->imag[k];
	    } else {
		hi[k] = 0.0;
	    }
	} else {
	    hr[k] = 0.0;
	    hi[k] = 0.0;
	}
    }
    fft(hr, hi, fftl, 0);

    /* loop for every block of data */
    for (i = 0; i < nloop; i++) {
	/* fft of input speech data */
	for (k = 0; k < fftl; k++) {
	    if (k < block && (pos = i * block + k) < x->length) {
		xr[k] = x->data[pos];
		if (x->imag != NULL) {
		    xi[k] = x->imag[pos];
		} else {
		    xi[k] = 0.0;
		}
	    } else {
		xr[k] = 0.0;
		xi[k] = 0.0;
	    }
	}
	fft(xr, xi, fftl, 0);

	/* multiplication in frequency domain */
	for (k = 0; k < fftl; k++) {  
	    real = xr[k] * hr[k] - xi[k] * hi[k];
	    imag = xr[k] * hi[k] + xi[k] * hr[k];
	    xr[k] = real;
	    xi[k] = imag;
	}

	/* ifft */
	fft(xr, xi, fftl, 1);

	/* overlap adding */
	for (k = 0; k < fftl && (pos = i * block + k) < xo->length; k++) {
	    if (pos >= 0) {
		xo->data[pos] += xr[k];
		if (xo->imag != NULL) {
		    xo->imag[pos] += xi[k];
		}
	    }
	}
    }

    /* memory free */
    free(hr);
    free(hi);
    free(xr);
    free(xi);

    return xo;
}

DVector xdvfftfilt(DVector b, DVector x, long fftl)
{
    int fftl_p;
    long i, nloop;
    long pos;
    long length, block;
    DVector sb;
    DVector sx;
    DVector xo;

    /* get fft point */
    if (x->length < b->length) {
	length = b->length + x->length - 1;
    } else {
	length = b->length;
    }
    fftl_p = nextpow2(MAX(fftl, 2 * length));
    fftl = POW2(fftl_p);
	
    /* initialize */
    block = fftl - b->length + 1;
    nloop = getnumloop(x->length, block);

    /* memory allocate */
    if (x->imag != NULL || b->imag != NULL) {
	/* xo->length is different from that of xdvfftfiltm */
	xo = xdvrizeros(x->length + b->length - 1);
    } else {
	xo = xdvzeros(x->length + b->length - 1);
    }

    /* fft of the filter */
    sb = xdvfft(b, fftl);

    /* loop for every block of data */
    for (i = 0; i < nloop; i++) {
	pos = i * block;

	/* fft of input speech data */
	sx = xdvrizeros(fftl);
	dvadd(sx, 0, x, pos, block, 0);
	dvfft(sx);

	/* multiplication in frequency domain */
	dvoper(sx, "*", sb);

	/* ifft */
	dvifft(sx);

	/* overlap adding */
	dvpaste(xo, sx, pos, sx->length, 1);

	/* memory free */
	xdvfree(sx);
    }

    /* memory free */
    xdvfree(sb);

    return xo;
}

DVector xdvfftfiltm2(DVector b, DVector x, long fftl)
{
    int fftl_p;
    long i, nloop;
    long pos;
    long length, block;
    DVector sb;
    DVector sx;
    DVector xo;

    /* get fft point */
    if (x->length < b->length) {
	length = b->length + x->length - 1;
    } else {
	length = b->length;
    }
    fftl_p = nextpow2(MAX(fftl, 2 * length));
    fftl = POW2(fftl_p);
	
    /* initialize */
    block = fftl - b->length + 1;
    nloop = getnumloop(x->length, block);

    /* memory allocate */
    if (x->imag != NULL || b->imag != NULL) {
	xo = xdvrizeros(x->length);
    } else {
	xo = xdvzeros(x->length);
    }

    /* fft of the filter */
    sb = xdvfft(b, fftl);

    /* loop for every block of data */
    for (i = 0; i < nloop; i++) {
	pos = i * block;

	/* fft of input speech data */
	sx = xdvrizeros(fftl);
	dvadd(sx, 0, x, pos, block, 0);
	dvfft(sx);

	/* multiplication in frequency domain */
	dvoper(sx, "*", sb);

	/* ifft */
	dvifft(sx);

	/* overlap adding */
	dvpaste(xo, sx, pos, sx->length, 1);

	/* memory free */
	xdvfree(sx);
    }

    /* memory free */
    xdvfree(sb);

    return xo;
}

DVector xdvconv(DVector a, DVector b)
{
    long i, j, pos;
    double ar, ai;
    double br, bi;
    double sum, sumi;
    DVector c;

    if (a->imag != NULL || b->imag != NULL) {
	c = xdvrizeros(a->length + b->length - 1);
    } else {
	c = xdvzeros(a->length + b->length - 1);
    }

    for (i = 0; i < c->length; i++) {
	if (c->imag != NULL) {
	    for (j = 0, sum = sumi = 0.0; j < a->length; j++) {
		pos = i - j;
		if (pos >= 0 && pos < b->length) {
		    ar = a->data[j];
		    br = b->data[pos];
		    if (a->imag == NULL) {
			ai = 0.0;
		    } else {
			ai = a->imag[j];
		    }
		    if (b->imag == NULL) {
			bi = 0.0;
		    } else {
			bi = b->imag[pos];
		    }
		    sum += ar * br - ai * bi;
		    sumi += ar * bi + ai * br;
		}
	    }
	    c->data[i] = sum;
	    c->imag[i] = sumi;
	} else {
	    for (j = 0, sum = 0.0; j < a->length; j++) {
		pos = i - j;
		if (pos >= 0 && pos < b->length) {
		    sum += a->data[j] * b->data[pos];
		}
	    }
	    c->data[i] = sum;
	}
    }

    return c;
}

DVector xdvfilter(DVector b, DVector a, DVector x)
{
    long i, j, pos;
    double ar, ai;
    double br, bi;
    double xr, xi;
    double arsum, aisum;
    double brsum, bisum;
    double value;
    DVector xo;

    if (a->imag != NULL || b->imag != NULL || x->imag != NULL) {
	xo = xdvrizeros(x->length);
    } else {
	xo = xdvzeros(x->length);
    }

    for (i = 0; i < xo->length; i++) {
	if (xo->imag != NULL) {
	    for (j = 1, arsum = aisum = 0.0; j < a->length; j++) {
		pos = i - j;
		if (pos >= 0 && pos < xo->length) {
		    ar = a->data[j];
		    xr = xo->data[pos];
		    if (a->imag == NULL) {
			ai = 0.0;
		    } else {
			ai = a->imag[j];
		    }
		    if (xo->imag == NULL) {
			xi = 0.0;
		    } else {
			xi = xo->imag[pos];
		    }
		    arsum += ar * xr - ai * xi;
		    aisum += ar * xi + ai * xr;
		}
	    }

	    for (j = 0, brsum = bisum = 0.0; j < b->length; j++) {
		pos = i - j;
		if (pos >= 0 && pos < x->length) {
		    br = b->data[j];
		    xr = x->data[pos];
		    if (b->imag == NULL) {
			bi = 0.0;
		    } else {
			bi = b->imag[j];
		    }
		    if (x->imag == NULL) {
			xi = 0.0;
		    } else {
			xi = x->imag[pos];
		    }
		    brsum += br * xr - bi * xi;
		    bisum += br * xi + bi * xr;
		}
	    }

	    if (a->imag == NULL || a->imag[0] == 0.0) {
		xo->data[i] = (brsum - arsum) / a->data[0];
		xo->imag[i] = (bisum - aisum) / a->data[0];
	    } else {
		value = CSQUARE(a->data[0], a->imag[0]);
		xr = brsum - arsum;
		xi = bisum - aisum;
		xo->data[i] = (xr * a->data[0] + xi * a->imag[0]) / value;
		xo->imag[i] = (-xr * a->imag[0] + xi * a->data[0]) / value;
	    }
	} else {
	    for (j = 1, arsum = 0.0; j < a->length; j++) {
		pos = i - j;
		if (pos >= 0 && pos < xo->length) {
		    arsum += a->data[j] * xo->data[pos];
		}
	    }

	    for (j = 0, brsum = 0.0; j < b->length; j++) {
		pos = i - j;
		if (pos >= 0 && pos < x->length) {
		    brsum += b->data[j] * x->data[pos];
		}
	    }

	    xo->data[i] = (brsum - arsum) / a->data[0];
	}
    }

    return xo;
}

#if 1
int convfile(char *a_filename, char *b_filename, int i_swap, char *o_filename, int o_swap,
	     double coef)
{
    long k, l;
    FILE *a_fp, *b_fp, *o_fp;
    int nwrite;
    long pos;
    long num_skip;
    long a_length, b_length, c_length;
    double a, b, c;

    a_length = getsiglen(a_filename, 0, short);
    b_length = getsiglen(b_filename, 0, short);
    
    /* open file */
    if (NULL == (a_fp = spOpenFile(a_filename, "rb"))) {
	return SP_FAILURE;
    }

    /* open file */
    if (NULL == (b_fp = spOpenFile(b_filename, "rb"))) {
	return SP_FAILURE;
    }

    /* open file */
    if (NULL == (o_fp = spOpenFile(o_filename, "wb"))) {
	return SP_FAILURE;
    }

    c_length = a_length + b_length - 1;
    
    for (k = 0; k < c_length; k++) {
	/*printf("k = %ld\n", k);*/
	
	if (k > 0) {
	    fseek(a_fp, 0, 0);
	    fseek(b_fp, 0, 0);
	    num_skip = fskipdata(MIN(k, b_length), b_fp, short);
	}

	c = 0.0;
	for (l = 0; l < a_length; l++) {
	    freadshorttod(&a, 1, i_swap, a_fp);
	    
	    pos = k - l;
	    if (pos >= 0 && pos < b_length) {
		fseek(b_fp, 0, 0);
		num_skip = fskipdata(pos, b_fp, short);
		freadshorttod(&b, 1, i_swap, b_fp);
		c += a * b;
		/*printf("c = %f, a = %f, b = %f\n", c, a, b);*/
	    }
	}

	if (coef != 0.0) c *= coef;

	nwrite = fwritedoubletos(&c, 1, o_swap, o_fp);
    }
    
    spCloseFile(a_fp);
    spCloseFile(b_fp);
    spCloseFile(o_fp);

    return SP_SUCCESS;
}
#endif
