/*
 *	input_aiff.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>

#include <sp/spInputPluginP.h>
#include <sp/spPluginMain.h>

#include <sp/spWave.h>

#include "plugin_aiff.c"

static char *sp_aiff_file_type_list[] = {
    SP_WAVE_AIFF_ID,
    SP_WAVE_AIFC_ID,
    SP_WAVE_AIFC_ID,
    SP_WAVE_AIFC_ID,
    NULL,
};
static char *sp_aiff_file_desc_list[] = {
    SP_AIFF_TYPE_NONE_STRING,
    SP_AIFF_TYPE_PCM_STRING,
    SP_AIFF_TYPE_ULAW_STRING,
    SP_AIFF_TYPE_ALAW_STRING,
    NULL,
};
static char *sp_aiff_file_filter_list[] = {
    "*.aif",
    "*.afc",
    "*.afc",
    "*.afc",
    NULL,
};
    
static spBool spIsSupportedByPluginAiff(char *filename);

static spPluginError spOpenPluginAiff(void *instance, char *filename, char *mode);
static spBool spClosePluginAiff(void *instance);
static spPluginState spGetPluginStateAiff(void *instance);
static long spGetPluginCurrentPositionAiff(void *instance);
static spBool spStopPluginAiff(void *instance);
static spBool spPausePluginAiff(void *instance);
static spBool spRestartPluginAiff(void *instance);

static long spReadPluginAiff(void *instance, char *data, long length);
static spBool spSeekPluginAiff(void *instance, long pos);
static long spGetPluginTotalLengthAiff(void *instance);


static spInputPluginRec sp_input_plugin_aiff = {
    NULL,
    NULL,

    SP_PLUGIN_INPUT,
    "AIFF",
    2,
    SP_PLUGIN_PRIORITY_MIDDLE,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginAiff,
    spFreePluginAiff,
    "AIFF Input Plugin",
    "AIFF Input Plugin Version 0.2\nCopyright(C) Hideki Banno\nE-mail: banno@itakura.nuee.nagoya-u.ac.jp",
    
    spInitPluginInstanceAiff,
    spFreePluginInstanceAiff,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_aiff_file_type_list,
    sp_aiff_file_desc_list,
    sp_aiff_file_filter_list,
    spIsSupportedByPluginAiff,
    spSetPluginFileTypeAiff,  
    spGetPluginFileTypeAiff,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK | SP_SONG_COPYRIGHT_MASK),
    spSetPluginSongInfoAiff,
    spGetPluginSongInfoAiff,
    spGetPluginBestSuffixAiff,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,

    spSetPluginSampleBitAiff,   
    spGetPluginSampleBitAiff,   
    spSetPluginChannelAiff,     
    spGetPluginChannelAiff,     
    spSetPluginSampleRateAiff,  
    spGetPluginSampleRateAiff,  
    spSetPluginOtherInfoAiff,   
    spGetPluginOtherInfoAiff,
    
    spOpenPluginAiff,              
    spClosePluginAiff,             
    spGetPluginStateAiff,          
    spGetPluginCurrentPositionAiff,
    spStopPluginAiff,              
    spPausePluginAiff,             
    spRestartPluginAiff,

    spReadPluginAiff,             
    spSeekPluginAiff,
    spGetPluginTotalLengthAiff,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_input_plugin_aiff;
}

static spBool spIsSupportedByPluginAiff(char *filename)
{
    return spIsAiffFile(filename);
}

static spPluginError spOpenPluginAiff(void *instance, char *filename, char *mode)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, "rb")) == NULL) {
	return SP_PLUGIN_ERROR_OPEN;
    }
    spDebug(10, "spOpenPluginAiff", "filename = %s\n", filename);

    /* read header */
    if (spReadAiffInfo(&pinstance->info, pinstance->fp) == SP_FALSE) {
	spCloseFile(pinstance->fp); pinstance->fp = NULL;
	return SP_PLUGIN_ERROR_FAILURE;
    }
    if (spSetPluginSampleBitAiff(instance, pinstance->info.samp_bit) == SP_FALSE) {
	spCloseFile(pinstance->fp); pinstance->fp = NULL;
	return SP_PLUGIN_ERROR_SAMP_BIT;
    }
    pinstance->type_index =
	spFindPluginRecFileTypeIndex((spIoPluginRec *)&sp_input_plugin_aiff, pinstance->info.file_desc);
    
    pinstance->current_pos = 0;

    /* read song info */
    spReadAiffSongInfo(&pinstance->info, &pinstance->song_info, pinstance->fp);
    fseek(pinstance->fp, pinstance->info.header_size, SEEK_SET);
    spDebug(10, "spOpenPluginAiff", "pinstance->info.header_size = %d\n", pinstance->info.header_size);
    
    spGetPluginTotalLengthAiff(instance);
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginAiff(void *instance)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    if (pinstance->fp != NULL) {
	/* close file */
	spCloseFile(pinstance->fp);
	pinstance->fp = NULL;
	pinstance->info.header_size = 0;
	pinstance->current_pos = 0;
    }

    return SP_TRUE;
}

static spPluginState spGetPluginStateAiff(void *instance)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static long spGetPluginCurrentPositionAiff(void *instance)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    return pinstance->current_pos;
}

static spBool spStopPluginAiff(void *instance)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    if (pinstance->fp != NULL) {
	fseek(pinstance->fp, pinstance->info.header_size, SEEK_SET);
    }
    pinstance->current_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginAiff(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginAiff(void *instance)
{

    return SP_TRUE;
}

/*
 * The data for spReadPluginAiff must be 16 bit data
 * regardless of argument for spSetPluginSampleBitAiff.
 */
static long spReadPluginAiff(void *instance, char *data, long length)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;    
    long nread;
    long pos;
    
    if (pinstance->fp == NULL) return -1;

    if ((nread = spReadAiffData(&pinstance->info, data, length, pinstance->fp)) < 0) {
	return -1;
    }
    pos = pinstance->current_pos + (nread / pinstance->info.num_channel);

    if (pos > pinstance->total_length) {
	pos = pinstance->total_length;
	nread = (pos - pinstance->current_pos) * pinstance->info.num_channel;
    }
    pinstance->current_pos = pos;
    
    return nread;
}

static spBool spSeekPluginAiff(void *instance, long pos)
{
    long k, nloop;
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    if (pinstance->fp == NULL) return SP_FALSE;

    fseek(pinstance->fp, pinstance->info.header_size, SEEK_SET);

    /* for a long file */
    nloop = (long)pinstance->info.num_channel * (long)(pinstance->info.samp_bit / 8);
    for (k = 0; k < nloop; k++) {
	fseek(pinstance->fp, pos, SEEK_CUR);
    }
    pinstance->current_pos = pos;
    
    return SP_TRUE;
}

static long spGetPluginTotalLengthAiff(void *instance)
{
    spAiffPluginInstance pinstance = (spAiffPluginInstance)instance;
    
    if (pinstance->total_length <= 0) {
	pinstance->total_length = pinstance->info.length;
    }

    return pinstance->total_length;
}

