/*
 *	input_ogg.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <vorbis/vorbisfile.h>

#include <sp/spInputPluginP.h>
#include <sp/spPluginMain.h>

#include "plugin_ogg.c"

static char *sp_ogg_file_type_list[] = {
    "ogg",
    NULL,
};
static char *sp_ogg_file_desc_list[] = {
    "Ogg Vorbis",
    NULL,
};
static char *sp_ogg_file_filter_list[] = {
    "*.ogg",
    NULL,
};
    
static spBool spIsSupportedByPluginOgg(char *filename);

static spPluginError spOpenPluginOgg(void *instance, char *filename, char *mode);
static spBool spClosePluginOgg(void *instance);
static spPluginState spGetPluginStateOgg(void *instance);
static long spGetPluginCurrentPositionOgg(void *instance);
static spBool spStopPluginOgg(void *instance);
static spBool spPausePluginOgg(void *instance);
static spBool spRestartPluginOgg(void *instance);

static long spReadPluginOgg(void *instance, char *data, long length);
static spBool spSeekPluginOgg(void *instance, long pos);
static long spGetPluginTotalLengthOgg(void *instance);


static spInputPluginRec sp_input_plugin_ogg = {
    NULL,
    NULL,

    SP_PLUGIN_INPUT,
    "OGG",
    1,
    SP_PLUGIN_PRIORITY_HIGHER,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginOgg,
    spFreePluginOgg,
    "Ogg Vorbis Input Plugin",
    "Ogg Vorbis Input Plugin Version 0.1\nCopyright(C) Hideki Banno\nE-mail: banno@itakura.nuee.nagoya-u.ac.jp",
    
    spInitPluginInstanceOgg,
    spFreePluginInstanceOgg,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_ogg_file_type_list,
    sp_ogg_file_desc_list,
    sp_ogg_file_filter_list,
    spIsSupportedByPluginOgg,
    spSetPluginFileTypeOgg,  
    spGetPluginFileTypeOgg,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK
     | SP_SONG_GENRE_MASK | SP_SONG_RELEASE_MASK | SP_SONG_COPYRIGHT_MASK
     | SP_SONG_TRACK_MASK | SP_SONG_SOURCE_MASK | SP_SONG_ALBUM_MASK | SP_SONG_SOFTWARE_MASK),
    spSetPluginSongInfoOgg,
    spGetPluginSongInfoOgg,
    spGetPluginBestSuffixOgg,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,

    spSetPluginSampleBitOgg,   
    spGetPluginSampleBitOgg,   
    spSetPluginChannelOgg,     
    spGetPluginChannelOgg,     
    spSetPluginSampleRateOgg,  
    spGetPluginSampleRateOgg,  
    spSetPluginOtherInfoOgg,   
    spGetPluginOtherInfoOgg,
    
    spOpenPluginOgg,              
    spClosePluginOgg,             
    spGetPluginStateOgg,          
    spGetPluginCurrentPositionOgg,
    spStopPluginOgg,              
    spPausePluginOgg,             
    spRestartPluginOgg,

    spReadPluginOgg,             
    spSeekPluginOgg,
    spGetPluginTotalLengthOgg,
};


static int sp_ogg_fseek(FILE *fp, ogg_int64_t offset, int mode)
{
    int ret = 0;
    ogg_int64_t shift, remain;

    if (offset > 0 && (mode == SEEK_CUR || mode == SEEK_SET)) {
	remain = offset;
	while (remain > 0) {
	    shift = MIN(32767, remain);
	    if ((ret = fseek(fp, shift, mode)) == -1) {
		break;
	    }
	    remain -= shift;
	    mode = SEEK_CUR;
	}
    } else {
	ret = fseek(fp, offset, mode);
    }
    
    return ret;
}

static ov_callbacks sp_ogg_callbacks = {
    (size_t (*)(void *, size_t, size_t, void *))fread,
    (int (*)(void *, ogg_int64_t, int))sp_ogg_fseek,
    (int (*)(void *))fclose,
    (long (*)(void *))ftell,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_input_plugin_ogg;
}

static spBool spIsSupportedByPluginOgg(char *filename)
{
    FILE *fp;
    OggVorbis_File vf;
    spBool flag = SP_FALSE;

    if ((fp = spOpenFile(filename, "rb")) != NULL) {
	vf.datasource = NULL;
	if (ov_test_callbacks(fp, &vf, NULL, 0, sp_ogg_callbacks) >= 0) {
	    ov_clear(&vf);
	    flag = SP_TRUE;
	} else if (vf.datasource == NULL) {
	    spCloseFile(fp);
	}
    }
    
    spDebug(10, "spIsSupportedByPluginOgg", "filename = %s: flag = %d\n", filename, flag);
    
    return flag;
}

static spBool getComments(spOggPluginInstance pinstance, spSongInfo *song_info)
{
#if 1
    vorbis_comment *vc;
    
    if ((vc = ov_comment(&pinstance->vf, -1)) != NULL) {
	int i, j;
	char *ostr;
	char *iptr, *optr;
	char *comment;
	char buf[SP_MAX_LINE];
	spConverter conv;

	if ((conv = spOpenConverter("UTF-8", NULL)) != NULL) {
	    for (i = 0; i < vc->comments; i++) {
		comment = vc->user_comments[i];
		spDebug(10, "spOpenPluginOgg", "comments[%d] = %s\n", i, comment);
		
		if (strvcaseeq(comment, "TRACKNUMBER=")) { /* track */
		    song_info->info_mask |= SP_SONG_TRACK_MASK;
		    song_info->track = atoi(comment + 12);
		} else {
		    for (j = 0; sp_ogg_song_info_table[j].mask; j++) {
			sprintf(buf, "%s=", sp_ogg_song_info_table[j].name);
			
			if (strvcaseeq(comment, buf)) {
			    song_info->info_mask |= sp_ogg_song_info_table[j].mask;
			    optr = (char *)song_info + sp_ogg_song_info_table[j].offset;
			    
			    iptr = comment + strlen(buf);

			    if ((ostr = xspConvert(conv, iptr)) != NULL) {
				spStrCopy(optr, SP_SONG_INFO_SIZE, ostr);
				xfree(ostr);
			    }
			}
		    }
		}
	    }

#if 0
	    spShowSongInfo(song_info, NULL, stdout);
#endif
	    
	    spCloseConverter(conv);
	    return SP_TRUE;
	}
    }
#endif

    return SP_FALSE;
}

static spPluginError spOpenPluginOgg(void *instance, char *filename, char *mode)
{
    vorbis_info *vi;
    spOggPluginInstance pinstance = (spOggPluginInstance)instance;

    spInitSongInfo(&pinstance->song_info);
    memset(&pinstance->vf, 0, sizeof(OggVorbis_File));
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, "rb")) == NULL) {
	return SP_PLUGIN_ERROR_OPEN;
    }
    spDebug(10, "spOpenPluginOgg", "filename = %s\n", filename);

    if (ov_open_callbacks(pinstance->fp, &pinstance->vf, NULL, 0, sp_ogg_callbacks) < 0) {
	ov_clear(&pinstance->vf);
	
	/* fclose done by ov_clear if vf->datasource is not NULL */
	if (pinstance->vf.datasource == NULL) {
	    spCloseFile(pinstance->fp);
	}
	pinstance->fp = NULL;
	return SP_PLUGIN_ERROR_OPEN;
    }

    if ((vi = ov_info(&pinstance->vf, -1)) != NULL) {
	pinstance->num_channel = vi->channels;
	pinstance->samp_rate = vi->rate;
    }
    pinstance->total_length = ov_pcm_total(&pinstance->vf, -1);;
    pinstance->current_pos = 0;

    getComments(pinstance, &pinstance->song_info);

    spDebug(10, "spOpenPluginOgg", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginOgg(void *instance)
{
    spOggPluginInstance pinstance = (spOggPluginInstance)instance;
    
    if (pinstance->fp != NULL) {
	ov_clear(&pinstance->vf);
	
	/* fclose done by ov_clear */
	pinstance->fp = NULL;
	pinstance->current_pos = 0;
    }

    return SP_TRUE;
}

static spPluginState spGetPluginStateOgg(void *instance)
{
    spOggPluginInstance pinstance = (spOggPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static long spGetPluginCurrentPositionOgg(void *instance)
{
    spOggPluginInstance pinstance = (spOggPluginInstance)instance;
    
    return pinstance->current_pos;
}

static spBool spStopPluginOgg(void *instance)
{
    spOggPluginInstance pinstance = (spOggPluginInstance)instance;

    if (!ov_seekable(&pinstance->vf)) {
	return SP_FALSE;
    }
    ov_pcm_seek(&pinstance->vf, 0);
    pinstance->current_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginOgg(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginOgg(void *instance)
{

    return SP_TRUE;
}

static long spReadPluginOgg(void *instance, char *data, long length)
{
    spOggPluginInstance pinstance = (spOggPluginInstance)instance;    
    long nread;
    long pos;
    int current_section;
    
    if (pinstance->fp == NULL) return -1;

    spDebug(10, "spReadPluginOgg", "length = %ld\n", length);
    
    if ((nread = ov_read(&pinstance->vf, data, length * 2,
			 BYTE_ORDER == BIG_ENDIAN, 2, 1, &current_section)) < 0) {
	spDebug(10, "spReadPluginOgg", "read error: nread = %ld, section = %d\n", nread, current_section);
	return -1;
    }
    nread /= 2;
    spDebug(10, "spReadPluginOgg", "nread = %ld, section = %d\n", nread, current_section);

    pos = pinstance->current_pos + nread / pinstance->num_channel;
    pinstance->current_pos = pos;
    spDebug(10, "spReadPluginOgg", "pos = %ld\n", pos);
    
    return nread;
}

static spBool spSeekPluginOgg(void *instance, long pos)
{
    spOggPluginInstance pinstance = (spOggPluginInstance)instance;
    
    if (pinstance->fp == NULL) return SP_FALSE;
    
    if (!ov_pcm_seek(&pinstance->vf, pos)) {
	return SP_FALSE;
    }
    
    pinstance->current_pos = pos;
    
    return SP_TRUE;
}

static long spGetPluginTotalLengthOgg(void *instance)
{
    spOggPluginInstance pinstance = (spOggPluginInstance)instance;
    
    return pinstance->total_length;
}

