/*
 *	swWave.h
 */

#ifndef __SWWAVE_H
#define __SWWAVE_H

#include <sp/spDefs.h>
#include <sp/spThread.h>
#include <sp/spPlugin.h>
#include <sp/spOutputPlugin.h>
#include <sp/spInputPlugin.h>

#ifdef __cplusplus
extern "C" {
#endif

#define SW_DEFAULT_TEMP_FILE_SUFFIX ".ad"
#define SW_PEAK_FILE_SUFFIX ".spk"
#define SW_DEFAULT_PEAK_THIN_LENGTH 256
    
/* Callback procedure
 *
 * edit_func [pos = SW_PROCESS_STARTED] --> edit_func [pos = current position]
 * --> edit_finish_func --> edit_func [pos = SW_PROCESS_FINISHED]
 *
 * play_func [pos = SW_PROCESS_STARTED] --> play_func [pos = current position]
 * --> play_func [pos = SW_PROCESS_FINISHED]
 *
 * swIsWaveProcessing will be SP_FALSE when pos = SW_PROCESS_FINISHED
 */

#define SW_USE_THREAD

#define SW_ERROR_PLAY_LONG SP_PLUGIN_ERROR_USER
#define SW_ERROR_LOOP_PLAY_SHORT (SP_PLUGIN_ERROR_USER+1)
#define SW_ERROR_DIFFERENT_CHANNEL (SP_PLUGIN_ERROR_USER+2)
#define SW_ERROR_DIFFERENT_SAMP_RATE (SP_PLUGIN_ERROR_USER+3)
#define SW_ERROR_SOURCE_WAVE_NOT_FOUND (SP_PLUGIN_ERROR_USER+4)
#define SW_ERROR_SOURCE_WAVE_PROCESSING (SP_PLUGIN_ERROR_USER+5)

#define SW_PROCESS_FINISHED -1
#define SW_PROCESS_ERROR -2
#define SW_PROCESS_STARTED -3

#define SW_MIN_LOOP_PLAY_LENGTH 8192
    
typedef enum {
    SW_TIME_FORMAT_UNKNOWN = -1,
    SW_TIME_FORMAT_SEC = 0,
    SW_TIME_FORMAT_MSEC = 1,
    SW_TIME_FORMAT_POINT = 2,
    SW_TIME_FORMAT_SEPARATED_SEC = 3,
    SW_TIME_FORMAT_FLOORED_MSEC = 4,
    SW_TIME_FORMAT_INDEXED = 5,
    SW_TIME_FORMAT_SEC_WITH_DATA = 6,
    SW_TIME_FORMAT_MSEC_WITH_DATA = 7,
    SW_TIME_FORMAT_POINT_WITH_DATA = 8,
    SW_TIME_FORMAT_SEPARATED_SEC_WITH_DATA = 9,
    SW_TIME_FORMAT_FLOORED_MSEC_WITH_DATA = 10,
    SW_TIME_FORMAT_INDEXED_WITH_DATA = 11,
    SW_TIME_FORMAT_ESPS = 12,
} swTimeFormat;

typedef enum {
    SW_EDIT_NONE = -1,
    SW_EDIT_CROP,
    SW_EDIT_EXTRACT,
    SW_EDIT_DELETE,
    SW_EDIT_ERASE,
    SW_EDIT_COPY,
    
    SW_EDIT_PASTE,
    SW_EDIT_INSERT,
    SW_EDIT_MIX,
    SW_EDIT_REPLACE,
    
    SW_EDIT_AMPLIFY,
    SW_EDIT_FADE_IN,
    SW_EDIT_FADE_OUT,
    SW_EDIT_CHANNEL_SWAP,
    SW_EDIT_BIT_CONV,
    SW_EDIT_SAMP_RATE_CONV,
    SW_EDIT_CHANNEL_CONV,
    SW_EDIT_MONAURALIZE,
    SW_EDIT_CHANGE_VALUE,
    
    SW_EDIT_READ,
    SW_EDIT_WRITE,
    
    SW_EDIT_INSERT_PAUSE,
    SW_EDIT_LONG_TERM_ANALYSIS,
} swEditType;

typedef enum {
    SW_ANALYSIS_NONE = -1,
    SW_ANALYSIS_SPECTRUM = 0,
    SW_ANALYSIS_SMOOTHED_SPECTRUM = 1,
    SW_ANALYSIS_PHASE = 2,
    SW_ANALYSIS_UNWRAPPED_PHASE = 3,
    SW_ANALYSIS_GROUP_DELAY = 4,
    SW_ANALYSIS_CEPSTRUM = 5,
} swAnalysisType;

typedef enum {
    SW_WINDOW_NONE = -1,
    SW_WINDOW_RECTANGLE = 0,
    SW_WINDOW_HAMMING = 1,
    SW_WINDOW_HANNING = 2,
    SW_WINDOW_BLACKMAN = 3,
    SW_WINDOW_GAUSS = 4,
} swWindowType;

typedef struct _swWave *swWave;
typedef spBool (*swWaveErrorCallback)(swWave wave, spBool in_thread, int error, void *data);
typedef spBool (*swWaveOptionCallback)(swWave wave, spBool in_thread, spOptions options, void *data);
typedef spBool (*swWavePlayCallback)(swWave wave, spBool in_thread, long pos, void *data);
typedef spBool (*swWaveEditCallback)(swWave wave, spBool in_thread, long pos, swEditType edit_type, void *data);
typedef spBool (*swWaveEditFinishCallback)(swWave wave, swWave owave, spBool in_thread, swEditType edit_type, void *data);
typedef spBool (*swWaveIoCallback)(swWave wave, spBool in_thread, long pos, void *data);

typedef struct _swLabel swLabel;
typedef struct _swLabels *swLabels;

typedef struct _swProcessConfig  swProcessConfig;
    
struct _swProcessConfig {
    swWave wave;
    swEditType edit_type;
    int samp_bit;
    long offset;
    long length;
    double value;
    spBool flag;
    void *data;
};

typedef struct _swWaveConfig {
    long max_read_length;
    long read_callback_length;
    long normal_read_length;
    long play_read_length;
    long buffer_length;
    int max_num_undo;
    
    spBool thread_safe;
    spBool play_use_audio;
    spBool play_use_wav;
    spBool process_use_thread;
    spBool linear_spectrum;
    spBool normalize_spectrum;
    spBool float_normalized;
    
    long audio_buffer_size;
    char temp_dir[SP_MAX_PATHNAME];

    swTimeFormat label_format;
    swTimeFormat region_label_format;
    
    long sfc_buffer_length;
    double sfc_cutoff;
    double sfc_sidelobe;
    double sfc_transition;
    double sfc_tolerance;
    double sfc_gain;
    
    swWaveErrorCallback error_func;
    swWaveOptionCallback option_func;
    swWavePlayCallback play_func;
    swWaveEditCallback edit_func;
    swWaveEditFinishCallback edit_finish_func;
    swWaveIoCallback read_func;
} swWaveConfigRec, *swWaveConfig;
    
struct _swWave {
    char *plugin_name;
    char *filename;
    char *file_type;
    char *orig_plugin_name;
    char *orig_filename;
    char *orig_file_type;
    char *new_plugin_name;
    char *new_filename;
    char *new_file_type;
    int num_channel;
    int selected_channel;	/* -1: not selected */
    int samp_bit;
    double samp_rate;
    double exact_samp_rate;
    unsigned long song_info_mask;
    spSongInfo song_info;
    long offset;		/* offset position in file */
    long length;		/* buffer length */
    long data_length;		/* data length (!= length) */
    long total_length;		/* total data length */
    long thin_length;		/* data thinned out length */
    long alloc_size;		/* allocated buffer size */
    spBool float_flag;
    spBool orig_flag;
    spBool edit_flag;
    spBool overflow;
    spBool thread_safe;
    char *data;
    short min;
    short max;
    long lmin;
    long lmax;
    double dmin;
    double dmax;
    spBool range_available;

    char *peak_filename;
    long peak_length;
    long peak_thin_length;

    long peak_buf_alloc_size;
    long peak_buf_length;
    long peak_buf_thin_length;
    char *peak_buf;

    long buf_alloc_size;
    long buf_length;
    long buf_load_length;
    char *buf;

    long readbuf_size;
    char *readbuf;
    long editbuf_length;
    char *editbuf;

    long edit_offset;
    long edit_length;
    long play_offset;
    long play_length;

    spBool process_finished;
    spBool process_flag;
    spBool play_flag;
    void *call_data;

    swProcessConfig process_config;
    
    void *edit_thread;
    void *audio_thread;
    void *mutex;
    void *io_mutex;
    
    swLabels labels;
    
    swWaveConfig config;

    char *edit_name;
    swWave prev_wave;
    swWave next_wave;
};

#if defined(MACOS)
#pragma import on
#endif

extern spBool swIsWaveNone(swWave wave);
extern spBool swIsWaveFloat(swWave wave);
extern spBool swIsWaveLong(swWave wave);
extern spBool swIsWaveOverflow(swWave wave);
extern spBool swIsWavePlaying(swWave wave);
extern spBool swIsWaveProcessing(swWave wave);
extern spBool swIsWaveProcessFinished(swWave wave);
extern spBool swIsWaveEdited(swWave wave);
extern spBool swIsWaveThreadSafe(swWave wave);
extern spBool swIsWavePeakAvailable(swWave wave);
extern spBool swNeedProcessBlock(swWave src_wave, swWave wave);

extern double swGetWaveSampleRate(swWave wave);
extern double swGetWaveExactSampleRate(swWave wave);
extern int swGetWaveSampleBit(swWave wave);
extern int swGetWaveNumChannel(swWave wave);
extern int swGetWaveSelectedChannel(swWave wave);

extern long swGetWaveOffset(swWave wave);
extern long swGetWaveLength(swWave wave);
extern long swGetWaveEditOffset(swWave wave);
extern long swGetWaveEditLength(swWave wave);
extern long swGetWaveTotalLength(swWave wave);
extern long swGetWaveThinLength(swWave wave);
extern spBool swSetWavePlayRegion(swWave wave, long offset, long length);

extern char *swGetWaveFileName(swWave wave);
extern char *swGetWaveOriginalFileName(swWave wave);
extern char *swGetWaveFileType(swWave wave);
extern char *swGetWavePluginName(swWave wave);
extern spSongInfo *swGetWaveSongInfo(swWave wave);
extern swLabels swGetWaveLabels(swWave wave);
extern char *swGetEditName(swWave wave);
extern spBool swSetEditName(swWave wave, char *edit_name);
extern swWaveConfig swGetWaveConfig(swWave wave);

extern long swConvertLengthToByte(swWave wave, long length);
extern long swConvertByteToLength(swWave wave, long size);
extern long swLoadIntoBuffer(spPlugin *plugin, swWave wave, int channel, long length, char *buf);
    

extern spBool swSetTempDir(swWaveConfig config, char *dir);
extern char *swGetTempFile(swWaveConfig config, char *suffix);

extern spBool swIsOverflow(int samp_bit, double value);
extern double swGetLimitValue(int samp_bit);
extern spBool swAllocData(swWave wave, long alloc_length);
extern spBool swAllocBuffer(swWave wave, long alloc_length);
extern spBool swInitWaveRange(swWave wave);
extern swWave swInitWave(swWaveConfig config, char *filename, char *plugin_name,  char *file_type,
			 int samp_bit, int num_channel, double samp_rate, spBool orig_flag);
extern void _swFreeWave(swWave wave);
extern void _swDestroyWave(swWave wave);
extern unsigned long swGetSongInfoMask(swWave wave);
extern swWave swGetWave(swWaveConfig config, char *filename, char *plugin_name, char *file_type,
			int samp_bit, int num_channel, double samp_rate,
			spBool orig_flag, spBool callback_flag, spPluginError *error);
extern spPlugin *swOpenWave(swWave wave, char *mode);
extern long swReadWave(swWave wave, spBool in_thread, long offset, long length);
extern long swReadTotalWave(swWave wave, spBool in_thread);
extern long swReadBuffer(spPlugin *plugin, swWave wave, int channel, long length);
extern long swWriteData(spPlugin *plugin, swWave wave);
extern spBool swWriteWave(char *filename, char *plugin_name, char *file_type,
			  swWave wave, spBool callback_flag, spBool use_thread);
extern swWave swGetCroppedWave(char *filename, swWave wave, long offset, long length, spBool in_thread);
extern spBool swSetPrevWave(swWave wave, swWave prev_wave);
extern spBool swCanUndoWave(swWave wave);
extern spBool swCanRedoWave(swWave wave);
extern spBool swUndoWave(swWave *wave);
extern spBool swRedoWave(swWave *wave);
extern spBool swIsWaveRangeAvailable(swWave wave);
extern spBool swSetWaveRange(swWave wave, double min, double max);
extern double swGetWaveMax(swWave wave);
extern double swGetWaveMin(swWave wave);
extern spBool swGetBufferRange(swWave wave, long channel, double *minp, double *maxp,
			       long *min_offsetp, long *max_offsetp);
extern spBool swGetWaveBufferData(swWave wave, long channel, long pos, long buf_length, char *buf, double *value);
extern spBool swSetWaveBufferData(swWave wave, long channel, long pos, long buf_length, char *buf, double value);
extern spBool swGetWaveData(swWave wave, long channel, long pos, double *value);
extern spBool swSetWaveData(swWave wave, long channel, long pos, double value);
extern double swGetBufferData(swWave wave, long channel, long pos);
extern spBool swSetBufferData(swWave wave, long channel, long pos, double value);
extern spBool swGetPeakData(swWave wave, long channel, long pos, double *value);

extern void *swCreateProcessThread(int priority, spThreadFunc func, swProcessConfig *config);
extern spBool swLockMutex(swWave wave);
extern spBool swUnlockMutex(swWave wave);
extern spBool swWaitMutex(swWave wave);
    
extern swWave swExtractWave(swWave wave, long offset, long length);
extern swWave swCropWave(swWave wave, long offset, long length);
extern swWave swDeleteWave(swWave wave, long offset, long length);
extern swWave swEraseWave(swWave wave, long offset, long length);
extern swWave swAmplifyWave(swWave wave, long offset, long length, double value);
extern swWave swFadeInWave(swWave wave, long offset, long length);
extern swWave swFadeOutWave(swWave wave, long offset, long length);
extern swWave swSwapWaveChannel(swWave wave, long offset, long length);
extern swWave swConvertWaveBit(swWave wave, int samp_bit);
extern swWave swConvertWaveSampFreq(swWave wave, double samp_rate);
extern swWave swMonauralizeWave(swWave wave);
extern swWave swChangeWaveValue(swWave wave, int channel, long offset, double value);

extern swWave swPasteWave(swWave wave, swWave src_wave, long offset, int channel);
extern swWave swMixWave(swWave wave, swWave src_wave, long offset, int channel);
extern swWave swInsertWave(swWave wave, swWave src_wave, long offset, int channel);
extern swWave swInsertPauseWave(swWave wave, long offset, long pause_length);
extern swWave swReplaceWave(swWave wave, swWave src_wave, long offset, long length);

extern spBool swSetWaveErrorCallback(swWaveConfig config, swWaveErrorCallback func);
extern spBool swSetWaveOptionCallback(swWaveConfig config, swWaveOptionCallback func);
extern spBool swSetWavePlayCallback(swWaveConfig config, swWavePlayCallback func);
extern spBool swSetWaveReadCallback(swWaveConfig config, swWaveIoCallback func);
extern spBool swSetWaveEditCallback(swWaveConfig config, swWaveEditCallback func);
extern spBool swSetWaveEditFinishCallback(swWaveConfig config, swWaveEditFinishCallback func);
extern spBool swSetWaveCallbackData(swWave wave, void *data);
extern spBool swProcessStop(swWave wave);

extern swWave swAnalysisWave(swWave iwave, long offset, long length,
			     swAnalysisType analysis_type, swWindowType window_type,
			     long fftl, double lifterm);

#define swFreeWave(wave) {_swFreeWave(wave);(wave)=NULL;}
#define swDestroyWave(wave) {_swDestroyWave(wave);(wave)=NULL;}

#if defined(MACOS)
#pragma import off
#endif

#define SP_WAVE_PLAY_DUMMY_TIME (0.5) /* [ms] */

#if 0
#ifndef _WIN32
#define ESPS
#endif
#endif
    
#define SP_WAVE_FORMAT_UNKNOWN_LABEL "Unknown"
#define SP_WAVE_FORMAT_LITTLE_LABEL "Little Endian"
#define SP_WAVE_FORMAT_BIG_LABEL "Big Endian"
#define SP_WAVE_FORMAT_ULAW_LABEL "U-law"
#define SP_WAVE_FORMAT_ALAW_LABEL "A-law"
#define SP_WAVE_FORMAT_TEXT_LABEL "Text"
#define SP_WAVE_FORMAT_TEXT_TIME_LABEL "Text with Time"
#define SP_WAVE_FORMAT_TEXT_FREQ_LABEL "Text with Frequency"
#define SP_WAVE_FORMAT_WAV_LABEL "Wav"
#ifdef ESPS
#define SP_WAVE_FORMAT_ESPS_LABEL "ESPS"
#endif
#define SP_WAVE_FORMAT_RAW_LABEL "Raw"
#define SP_WAVE_FORMAT_SWAP_LABEL "Swap Raw"

#ifdef __cplusplus
}  /* Close scope of 'extern "C"' declaration */
#endif

#endif /* __SWWAVE_H */
